@@ -40,7 +40,7 @@ use Mail::SpamAssassin::EncappedMessage;
 use Mail::Audit;
 
 use vars	qw{
-  	@ISA
+  	@ISA $base64alphabet
 };
 
 @ISA = qw();
@@ -57,6 +57,7 @@ sub new {
     'msg'		=> $msg,
     'hits'		=> 0,
     'test_logs'		=> '',
+    'test_names_hit'	=> '',
     'tests_already_hit' => { },
   };
 
@@ -75,23 +76,54 @@ sub check {
   # in order of slowness; fastest first, slowest last.
   # we do ALL the tests, even if a spam triggers lots of them early on.
   # this lets us see ludicrously spammish mails (score: 40) etc., which
-  # (TODO) we can then immediately submit to spamblocking services.
+  # we can then immediately submit to spamblocking services.
+  # Also, if parts of the message contain encoded bits (quoted-printable
+  # or base64), we test *both*.
 
   {
-    $self->{msg_body_array} = $self->{msg}->get_body();
-    $self->{full_msg_string} = $self->{msg}->get_all_headers()."\n".
-				  join ('', @{$self->{msg_body_array}});
-
     $self->do_head_tests();
-    $self->do_body_tests();
-    $self->do_body_eval_tests();
-    $self->do_full_tests();
-    $self->do_full_eval_tests();
-    $self->do_head_eval_tests();
 
-    # these are big, so delete them now.
-    delete $self->{msg_body_array};
-    delete $self->{full_msg_string};
+    # do body tests with raw text portions
+    {
+      my $bodytext = $self->get_raw_body_text_array();
+      $self->do_body_tests($bodytext);
+      $self->do_body_eval_tests($bodytext);
+      $bodytext = undef;
+    }
+
+    # do body tests with decoded portions
+    {
+      my $decoded = $self->get_decoded_body_text_array();
+      if (defined $decoded) {
+	$self->do_body_tests($decoded);
+	$self->do_body_eval_tests($decoded);
+      }
+      $decoded = undef;
+    }
+
+    # and do full tests: first with entire, full, undecoded message
+    {
+      my $fulltext = join ('', $self->{msg}->get_all_headers(), "\n",
+      				@{$self->{msg}->get_body()});
+      $self->do_full_tests(\$fulltext);
+      $self->do_full_eval_tests(\$fulltext);
+      $fulltext = undef;
+    }
+
+    # then with decoded message
+    {
+      my $decoded = $self->get_decoded_body_text_array();
+      if (defined $decoded) {
+	my $fulltext = join ('', $self->{msg}->get_all_headers(), "\n",
+				  @{$decoded});
+	$self->do_full_tests(\$fulltext);
+	$self->do_full_eval_tests(\$fulltext);
+	$fulltext = undef;
+      }
+      $decoded = undef;
+    }
+
+    $self->do_head_eval_tests();
   }
 
   $self->{required_hits} = $self->{conf}->{required_hits};
@@ -125,6 +157,53 @@ sub is_spam {
   return $self->{is_spam};
 }
 
+###########################################################################
+
+=item $list = $status->get_names_of_tests_hit ()
+
+After a mail message has been checked, this method can be called.  It will
+return a comma-separated string, listing all the symbolic test names
+of the tests which were trigged by the mail.
+
+=cut
+
+sub get_names_of_tests_hit {
+  my ($self) = @_;
+
+  $self->{test_names_hit} =~ s/,$//;
+  return $self->{test_names_hit};
+}
+
+###########################################################################
+
+=item $num = $status->get_hits ()
+
+After a mail message has been checked, this method can be called.  It will
+return the number of hits this message incurred.
+
+=cut
+
+sub get_hits {
+  my ($self) = @_;
+  return $self->{hits};
+}
+
+###########################################################################
+
+=item $num = $status->get_required_hits ()
+
+After a mail message has been checked, this method can be called.  It will
+return the number of hits required for a mail to be considered spam.
+
+=cut
+
+sub get_required_hits {
+  my ($self) = @_;
+  return $self->{required_hits};
+}
+
+###########################################################################
+
 =item $report = $status->get_report ()
 
 Deliver a "spam report" on the checked mail message.  This contains details of
@@ -157,8 +236,18 @@ The string C<*****SPAM*****> is prepended to the subject.
 
 =item X-Spam-Status: header for spam mails
 
-A string, C<Yes, hits=nn required=nn> is set in this header to reflect
-the filter status.
+A string, C<Yes, hits=nn required=nn tests=...> is set in this header to
+reflect the filter status.  The keys in this string are as follows:
+
+=over 4
+
+=item hits=nn The number of hits the message triggered.
+
+=item required=nn The threshold at which a mail is marked as spam.
+
+=item tests=... The symbolic names of tests which were triggered.
+
+=back
 
 =item X-Spam-Flag: header for spam mails
 
@@ -175,8 +264,9 @@ The SpamAssassin report is added to top of the mail message body.
 
 =item X-Spam-Status: header for non-spam mails
 
-A string, C<No, hits=nn required=nn> is set in this header to reflect
-the filter status.
+A string, C<No, hits=nn required=nn tests=...> is set in this header to reflect
+the filter status.  The keys in this string are the same as for spam mails (see
+above).
 
 =back
 
@@ -195,21 +285,53 @@ sub rewrite_mail {
 sub rewrite_as_spam {
   my ($self) = @_;
 
+  # message we'll be reading original values from. Normally the
+  # same as $self->{msg} (the target message for the rewritten
+  # mail), but if it already had spamassassin markup, we'll need
+  # to create a new $srcmsg to hold a 'cleaned-up' version.
+  my $srcmsg = $self->{msg};
+
+  if ($self->{msg}->get_header ("X-Spam-Status")) {
+    # the mail already has spamassassin markup. Remove it!
+    # bit messy this; we need to get the mail as a string,
+    # remove the spamassassin markup in it, then re-create
+    # a Mail::Audit object using a reference to the text 
+    # array (why not a string, ghod only knows).
+
+    my $text = $self->{main}->remove_spamassassin_markup ($self->{msg});
+    my @textary = split (/^/m, $text);
+    my %opts = ( 'data', \@textary );
+    my $audit = Mail::Audit->new (%opts);
+
+    # agh, we have to do this ourself?! why won't M::A do it right?
+    # for some reason it puts headers in the body
+    # while ($_ = shift @textary) { /^$/ and last; }
+    # $self->{msg}->replace_body (\@textary);
+
+    $srcmsg = $self->{main}->encapsulate_audit ($audit);
+
+    # delete the SpamAssassin-added headers in the target message.
+    $self->{msg}->delete_header ("X-Spam-Status");
+    $self->{msg}->delete_header ("X-Spam-Flag");
+    $self->{msg}->delete_header ("X-Spam-Prev-Content-Type");
+  }
+
   # First, rewrite the subject line.
-  $_ = $self->{msg}->get_header ("Subject"); $_ ||= '';
+  $_ = $srcmsg->get_header ("Subject"); $_ ||= '';
   s/^/\*\*\*\*\*SPAM\*\*\*\*\* /g;
   $self->{msg}->replace_header ("Subject", $_);
 
   # add some headers...
-  $_ = sprintf ("Yes, hits=%d required=%d",
-  			$self->{hits}, $self->{required_hits});
+
+  $_ = sprintf ("Yes, hits=%d required=%d tests=%s",
+	$self->{hits}, $self->{required_hits}, $self->get_names_of_tests_hit());
 
   $self->{msg}->put_header ("X-Spam-Status", $_);
   $self->{msg}->put_header ("X-Spam-Flag", 'YES');
 
   # defang HTML mail; change it to text-only.
-  my $ct = $self->{msg}->get_header ("Content-Type");
-  $ct ||= $self->{msg}->get_header ("Content-type");
+  my $ct = $srcmsg->get_header ("Content-Type");
+  $ct ||= $srcmsg->get_header ("Content-type");
 
   if (defined $ct && $ct ne '' && $ct ne 'text/plain') {
     $self->{msg}->replace_header ("Content-Type", "text/plain");
@@ -217,8 +339,9 @@ sub rewrite_as_spam {
     $self->{msg}->replace_header ("X-Spam-Prev-Content-Type", $ct);
   }
 
-  my $lines = $self->{msg}->get_body();
+  my $lines = $srcmsg->get_body();
   unshift (@{$lines}, split (/$/, $self->{report}));
+  ${$lines}[0] =~ s/\n//;
   $self->{msg}->replace_body ($lines);
 
   $self->{msg}->{audit};
@@ -227,8 +350,10 @@ sub rewrite_as_spam {
 sub rewrite_as_non_spam {
   my ($self) = @_;
 
-  $_ = sprintf ("No, hits=%d required=%d", $self->{hits},
-        $self->{required_hits});
+  $self->{test_names_hit} =~ s/,$//;
+
+  $_ = sprintf ("No, hits=%d required=%d tests=%s",
+	$self->{hits}, $self->{required_hits}, $self->{test_names_hit});
   $self->{msg}->put_header ("X-Spam-Status", $_);
   $self->{msg}->{audit};
 }
@@ -257,11 +382,21 @@ sub handle_auto_report {
 ###########################################################################
 # Non-public methods from here on.
 
-sub get_body_text {
+sub get_raw_body_text_array {
   my ($self) = @_;
   local ($_);
 
-  if (defined $self->{body_text}) { return $self->{body_text}; }
+  if (defined $self->{body_text_array}) { return $self->{body_text_array}; }
+
+  $self->{found_encoding_base64} = 0;
+  $self->{found_encoding_quoted_printable} = 0;
+
+  my $cte = $self->{msg}->get_header ('Content-Transfer-Encoding');
+  if (defined $cte && $cte =~ /quoted-printable/) {
+    $self->{found_encoding_quoted_printable} = 1;
+  } elsif (defined $cte && $cte =~ /base64/) {
+    $self->{found_encoding_base64} = 1;
+  }
 
   my $ctype = $self->{msg}->get_header ('Content-Type');
   $ctype ||=  $self->{msg}->get_header ('Content-type');
@@ -269,20 +404,28 @@ sub get_body_text {
 
   my $body = $self->{msg}->get_body();
   if ($ctype !~ /boundary="(.*)"/) {
-    $self->{body_text} = $body;
-    return $self->{body_text};
+    $self->{body_text_array} = $body;
+    return $self->{body_text_array};
   }
 
   # else it's a multipart MIME message. Skip non-text parts and
   # just assemble the body array from the text bits.
-  $self->{body_text} = [ ];
+  $self->{body_text_array} = [ ];
   my $multipart_boundary = "--$1\n";
   my $end_boundary = "--$1--\n";
 
   my @workingbody = @{$body};
 
   while ($_ = (shift @workingbody)) {
-    push (@{$self->{body_text}}, $_);
+    push (@{$self->{body_text_array}}, $_);
+
+    if (/^Content-Transfer-Encoding: /) {
+      if (/quoted-printable/) {
+	$self->{found_encoding_quoted_printable} = 1;
+      } elsif (/base64/) {
+	$self->{found_encoding_base64} = 1;
+      }
+    }
 
     if ($multipart_boundary eq $_) {
       $_ = (shift @workingbody);
@@ -291,12 +434,55 @@ sub get_body_text {
 
       # skip this attachment, it's non-text.
       while ($_ = (shift @workingbody)) {
-	last if ($multipart_boundary eq $_ || $end_boundary eq $_);
+	if ($end_boundary eq $_) { last; }
+	if ($multipart_boundary eq $_) { unshift (@workingbody, $_); last; }
       }
     }
   }
+  @workingbody = ();
 
-  return $self->{body_text};
+  return $self->{body_text_array};
+}
+
+###########################################################################
+
+sub get_decoded_body_text_array {
+  my ($self) = @_;
+  local ($_);
+  my $textary = $self->get_raw_body_text_array();
+
+  # TODO: doesn't yet handle checking multiple-attachment messages,
+  # where one part is qp and another is b64.  Instead the qp will
+  # be simply stripped.
+
+  if ($self->{found_encoding_base64}) {
+    $_ = '';
+    my $foundb64 = 0;
+    foreach my $line (@{$textary}) {
+      if (length($line) != 77) {	# 76 + newline
+	if ($foundb64) {
+	  $_ .= $line;		# last line of block is usually short
+	  last;
+	}
+      } else {
+	$_ .= $line; $foundb64 = 1;
+      }
+    }
+
+    $_ = $self->slow_base64_decode ($_);
+    # print "decoded: $_\n";
+    my @lines = split (/^/, $_);
+    return \@lines;
+
+  } elsif ($self->{found_encoding_quoted_printable}) {
+    $_ = join ('', $textary);
+    s/\=([0-9A-Fa-f]{2})/chr(hex($1))/ge; s/\=\n/\n/;
+    my @lines = split (/^/, $_);
+    return \@lines;
+
+  } else {
+    return undef;
+  }
 }
 
 ###########################################################################
@@ -305,7 +491,6 @@ sub get {
   my ($self, $hdrname, $defval) = @_;
   local ($_);
 
-
   if ($hdrname eq 'ALL') { return $self->{msg}->get_all_headers(); }
 
   my $getaddr = 0;
@@ -330,6 +515,9 @@ sub get {
   if ($getaddr) {
     s/^.*?<(.+)>\s*$/$1/g			# Foo Blah <jm@foo>
     	or s/^(.+)\s\(.*?\)\s*$/$1/g;	# jm@foo (Foo Blah)
+
+  } else {
+    $_ = $self->mime_decode_header ($_);
   }
 
   $_;
@@ -337,6 +525,60 @@ sub get {
 
 ###########################################################################
 
+# This function will decode MIME-encoded headers.  Note that it is ONLY
+# used from test functions, so destructive or mildly inaccurate results
+# will not have serious consequences.  Do not replace the original message
+# contents with anything decoded using this!
+#
+sub mime_decode_header {
+  my ($self, $enc) = @_;
+
+  # cf. http://www.nacs.uci.edu/indiv/ehood/MHonArc/doc/resources/charsetconverters.html
+
+  # quoted-printable encoded headers.
+  # ASCII:  =?US-ASCII?Q?Keith_Moore?= <moore@cs.utk.edu>
+  # Latin1: =?ISO-8859-1?Q?Keld_J=F8rn_Simonsen?= <keld@dkuug.dk>
+  # Latin1: =?ISO-8859-1?Q?Andr=E9_?= Pirard <PIRARD@vm1.ulg.ac.be>
+
+  if ($enc =~ s{=\?([^\?]+)\?Q\?([^\?]+)\?=}{
+    		$self->decode_mime_bit ($1, $2);
+	      }eg)
+  {
+    dbg ("decoded MIME header: \"$enc\"");
+  }
+
+  # TODO: handle base64-encoded headers. eg:
+  # =?UTF-8?B?Rlc6IFBhc3NpbmcgcGFyYW1ldGVycyBiZXR3ZWVuIHhtbHMgdXNp?=
+  # =?UTF-8?B?bmcgY29jb29uIC0gcmVzZW50IA==?=   (yuck)
+  # not high-priorty as they're still very rare.
+
+  return $enc;
+}
+
+sub decode_mime_bit {
+  my ($self, $encoding, $text) = @_;
+  local ($_) = $text;
+
+  if ($encoding =~ /^US-ASCII$/i
+  	|| $encoding =~ /^ISO-8859-\d+$/i
+  	|| $encoding =~ /^UTF-8$/i
+      )
+  {
+    # keep 8-bit stuff. forget mapping charsets though
+    s/_/ /g; s/\=([0-9A-Fa-f]{2})/chr(hex($1))/ge;
+  }
+
+  if ($encoding eq 'UTF-16')
+  {
+    # we just dump the high bits and keep the 8-bit chars.
+    s/_/ /g; s/=00//g; s/\=([0-9A-Fa-f]{2})/chr(hex($1))/ge;
+  }
+
+  return $_;
+}
+
+###########################################################################
+
 sub do_head_tests {
   my ($self) = @_;
   local ($_);
@@ -367,7 +609,7 @@ sub do_head_tests {
 }
 
 sub do_body_tests {
-  my ($self) = @_;
+  my ($self, $textary) = @_;
   my ($rulename, $pat);
   local ($_);
   $self->clear_test_state();
@@ -383,7 +625,6 @@ sub do_body_tests {
   }
 
   # generate the loop that goes through each line...
-  my $textary = $self->get_body_text();
   $evalstr = 'foreach $_ (@{$textary}) { study; '.$evalstr.'; }';
 
   # and run it.
@@ -394,7 +635,7 @@ sub do_body_tests {
 }
 
 sub do_full_tests {
-  my ($self) = @_;
+  my ($self, $fullmsgref) = @_;
   my ($rulename, $pat);
   local ($_);
   $self->clear_test_state();
@@ -405,13 +646,13 @@ sub do_full_tests {
   my $evalstr = '';
   while (($rulename, $pat) = each %{$self->{conf}->{full_tests}}) {
     $evalstr .= '
-      if ('.$pat.') { $self->got_hit (q{'.$rulename.'}, q{}); }
+      if ($$fullmsgref =~ '.$pat.') { $self->got_hit (q{'.$rulename.'}, q{}); }
     ';
   }
 
   # and run it.
-  $_ = $self->{full_msg_string};
-  if (!eval 'study; '.$evalstr.'; 1;') {
+  # study $$fullmsgref;
+  if (!eval $evalstr.'; 1;') {
     warn "Failed to run full SpamAssassin tests, skipping:\n".
 	      "\t($@)\n";
   }
@@ -425,13 +666,13 @@ sub do_head_eval_tests {
 }
 
 sub do_body_eval_tests {
-  my ($self) = @_;
-  $self->run_eval_tests ($self->{conf}->{body_evals}, 'BODY: ', $self->{msg_body_array});
+  my ($self, $bodystring) = @_;
+  $self->run_eval_tests ($self->{conf}->{body_evals}, 'BODY: ', $bodystring);
 }
 
 sub do_full_eval_tests {
-  my ($self) = @_;
-  $self->run_eval_tests ($self->{conf}->{full_evals}, '', $self->{full_msg_string});
+  my ($self, $fullmsgref) = @_;
+  $self->run_eval_tests ($self->{conf}->{full_evals}, '', $fullmsgref);
 }
 
 ###########################################################################
@@ -492,6 +733,8 @@ sub handle_hit {
   my $score = $self->{conf}->{scores}->{$rule};
   $self->{hits} += $score;
 
+  $self->{test_names_hit} .= $rule.",";
+
   $self->{test_logs} .= sprintf ("%-16s %s%s\n%s",
 		"Hit! (".$score." point".($score == 1 ? "":"s").")",
 		$area, $desc, $self->{test_log_msgs});
@@ -512,6 +755,57 @@ sub test_log {
   $self->{test_log_msgs} .= sprintf ("%16s [%s]\n", "", $msg);
 }
 
+###########################################################################
+# Rather than add a requirement for MIME::Base64, use a slower but
+# built-in base64 decode mechanism.
+#
+# original credit for this code:
+# b64decode -- decode a raw BASE64 message
+# A P Barrett <barrett@ee.und.ac.za>, October 1993
+# Minor mods by jm@jmason.org for spamassassin and "use strict"
+
+sub slow_base64_decode {
+  my ($self) = shift;
+  local ($_) = shift;
+
+  $base64alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.
+		    'abcdefghijklmnopqrstuvwxyz'.
+		    '0123456789+/'; # and '='
+
+  my $leftover = '';
+
+  # ignore illegal characters
+  s/[^$base64alphabet]//go;
+  # insert the leftover stuff from last time
+  $_ = $leftover . $_;
+  # if there are not a multiple of 4 bytes, keep the leftovers for later
+  m/^((....)*)/; $_=$&; $leftover=$';
+  # turn each group of 4 values into 3 bytes
+  s/(....)/&b64decodesub($1)/eg;
+  # special processing at EOF for last few bytes
+  if (eof) {
+      $_ .= &b64decodesub($leftover); $leftover = '';
+  }
+  # output it
+  return $_;
+}
+
+# b64decodesub -- takes some characters in the base64 alphabet and
+# returns the raw bytes that they represent.
+sub b64decodesub
+{
+  local ($_) = $_[0];
+	   
+  # translate each char to a value in the range 0 to 63
+  eval qq{ tr!$base64alphabet!\0-\77!; };
+  # keep 6 bits out of every 8, and pack them together
+  $_ = unpack('B*', $_); # look at the bits
+  s/(..)(......)/$2/g;   # keep 6 bits of every 8
+  s/((........)*)(.*)/$1/; # throw away spare bits (not multiple of 8)
+  $_ = pack('B*', $_);   # turn the bits back into bytes
+  $_; # return
+}
+
 ###########################################################################
 
 sub work_out_local_domain {
@@ -520,7 +814,6 @@ sub work_out_local_domain {
   # TODO -- if needed.
 
   # my @rcvd = $self->{msg}->get_header ("Received");
-  # print "JMD ".join (' ',@rcvd);
 
 # from dogma.slashnull.org (dogma.slashnull.org [212.17.35.15]) by
     # mail.netnoteinc.com (Postfix) with ESMTP id 3E010114097 for
